@@ -345,13 +345,12 @@ deriveInstCommon genericName repName gClass fromName toName opts n = do
   let origSigTy = if useKindSigs
                      then SigT origTy origKind
                      else origTy
-      tyIns = TySynInstD repName
-#if MIN_VERSION_template_haskell(2,9,0)
-                         (TySynEqn [origSigTy] tyInsRHS)
-#else
-                         [origSigTy] tyInsRHS
+  tyIns <- tySynInstDCompat repName
+#if MIN_VERSION_th_abstraction(0,3,0)
+                            Nothing
 #endif
-      ecOptions = emptyCaseOptions opts
+                            [return origSigTy] (return tyInsRHS)
+  let ecOptions = emptyCaseOptions opts
       mkBody maker = [clause [] (normalB $
         mkCaseExp gClass ecOptions name instTys cons maker) []]
       fcs = mkBody mkFrom
@@ -570,7 +569,7 @@ makeRepInline :: GenericClass
               -> Type
               -> Q Type
 makeRepInline gClass dv name instTys cons ty = do
-  let instVars = requiredTyVarsOfTypes [ty]
+  let instVars = freeVariablesWellScoped [ty]
       (tySynVars, gk)  = genericKind gClass instTys
 
       typeSubst :: TypeSubst
@@ -587,7 +586,7 @@ makeRepTySynApp gClass dv name ty =
   -- of the LHS of the Rep(1) instance. We call unKindedTV because the kind
   -- inferencer can figure out the kinds perfectly well, so we don't need to
   -- give anything here explicit kind signatures.
-  let instTvbs = map unKindedTV $ requiredTyVarsOfTypes [ty]
+  let instTvbs = map unKindedTV $ freeVariablesWellScoped [ty]
   in return $ applyTyToTvbs (genRepName gClass dv name) instTvbs
 
 -- | A backwards-compatible synonym for 'makeFrom0'.
@@ -659,8 +658,7 @@ repType :: GenericKind
         -> Q Type
 repType gk dv dt typeSubst cs =
     conT d1TypeName `appT` mkMetaDataType dv dt `appT`
-      foldr1' sum' (conT v1TypeName)
-        (map (repCon gk dv dt typeSubst) cs)
+      foldBal sum' (conT v1TypeName) (map (repCon gk dv dt typeSubst) cs)
   where
     sum' :: Q Type -> Q Type -> Q Type
     sum' a b = conT sumTypeName `appT` a `appT` b
@@ -708,9 +706,7 @@ repConWith :: GenericKind
            -> Q Type
 repConWith gk dv dt n typeSubst mbSelNames ssis ts isRecord isInfix = do
     let structureType :: Q Type
-        structureType = case ssis of
-                             [] -> conT u1TypeName
-                             _  -> foldr1 prodT f
+        structureType = foldBal prodT (conT u1TypeName) f
 
         f :: [Q Type]
         f = case mbSelNames of
@@ -795,7 +791,7 @@ mkCaseExp
   -> Q Exp
 mkCaseExp gClass ecOptions dt instTys cs matchmaker = do
   val <- newName "val"
-  lam1E (varP val) $ caseE (varE val) [matchmaker gClass ecOptions 1 0 dt instTys cs]
+  lam1E (varP val) $ caseE (varE val) [matchmaker gClass ecOptions 1 1 dt instTys cs]
 
 mkFrom :: GenericClass -> EmptyCaseOptions -> Int -> Int -> Name -> [Type]
        -> [ConstructorInfo] -> Q Match
@@ -807,8 +803,8 @@ mkFrom gClass ecOptions m i dt instTys cs = do
   where
     cases = case cs of
               [] -> errorFrom ecOptions dt
-              _  -> zipWith (fromCon gk wrapE (length cs)) [0..] cs
-    wrapE e = lrE m i e
+              _  -> zipWith (fromCon gk wrapE (length cs)) [1..] cs
+    wrapE e = lrE i m e
     (_, gk) = genericKind gClass instTys
 
 errorFrom :: EmptyCaseOptions -> Name -> [Q Match]
@@ -836,8 +832,8 @@ mkTo gClass ecOptions m i dt instTys cs = do
   where
     cases = case cs of
               [] -> errorTo ecOptions dt
-              _  -> zipWith (toCon gk wrapP (length cs)) [0..] cs
-    wrapP p = lrP m i p
+              _  -> zipWith (toCon gk wrapP (length cs)) [1..] cs
+    wrapP p = lrP i m p
     (_, gk) = genericKind gClass instTys
 
 errorTo :: EmptyCaseOptions -> Name -> [Q Match]
@@ -870,15 +866,10 @@ fromCon gk wrap m i
                    , constructorFields  = ts
                    }) = do
   checkExistentialContext cn vars ctxt
-  case ts of
-    [] -> match (conP cn [])
-                (normalB $ wrap $ lrE m i $ conE m1DataName `appE` (conE u1DataName)) []
-    _ -> do
-      fNames <- newNameList "f" $ length ts
-      match
-        (conP cn (map varP fNames))
-        (normalB $ wrap $ lrE m i $ conE m1DataName `appE`
-          foldr1 prodE (zipWith (fromField gk) fNames ts)) []
+  fNames <- newNameList "f" $ length ts
+  match (conP cn (map varP fNames))
+        (normalB $ wrap $ lrE i m $ conE m1DataName `appE`
+          foldBal prodE (conE u1DataName) (zipWith (fromField gk) fNames ts)) []
 
 prodE :: Q Exp -> Q Exp -> Q Exp
 prodE x y = conE productDataName `appE` x `appE` y
@@ -931,14 +922,9 @@ toCon gk wrap m i
                    , constructorFields  = ts
                    }) = do
   checkExistentialContext cn vars ctxt
-  case ts of
-    [] -> match (wrap $ lrP m i $ conP m1DataName [conP u1DataName []])
-                (normalB $ conE cn) []
-    _ -> do
-      fNames <- newNameList "f" $ length ts
-      match
-        (wrap $ lrP m i $ conP m1DataName
-          [foldr1 prod (zipWith (toField gk) fNames ts)])
+  fNames <- newNameList "f" $ length ts
+  match (wrap $ lrP i m $ conP m1DataName
+          [foldBal prod (conP u1DataName []) (zipWith (toField gk) fNames ts)])
         (normalB $ foldl appE (conE cn)
                          (zipWith (\nr -> resolveTypeSynonyms >=> toConUnwC gk nr)
                          fNames ts)) []
@@ -987,14 +973,20 @@ unboxRepName :: Type -> Name
 unboxRepName = maybe unK1ValName trd3 . unboxedRepNames
 
 lrP :: Int -> Int -> (Q Pat -> Q Pat)
-lrP 1 0 p = p
-lrP _ 0 p = conP l1DataName [p]
-lrP m i p = conP r1DataName [lrP (m-1) (i-1) p]
+lrP i n p
+  | n == 0       = fail "lrP: impossible"
+  | n == 1       = p
+  | i <= div n 2 = conP l1DataName [lrP i     (div n 2) p]
+  | otherwise    = conP r1DataName [lrP (i-m) (n-m)     p]
+                     where m = div n 2
 
 lrE :: Int -> Int -> (Q Exp -> Q Exp)
-lrE 1 0 e = e
-lrE _ 0 e = conE l1DataName `appE` e
-lrE m i e = conE r1DataName `appE` lrE (m-1) (i-1) e
+lrE i n e
+  | n == 0       = fail "lrE: impossible"
+  | n == 1       = e
+  | i <= div n 2 = conE l1DataName `appE` lrE i     (div n 2) e
+  | otherwise    = conE r1DataName `appE` lrE (i-m) (n-m)     e
+                     where m = div n 2
 
 unboxedRepNames :: Type -> Maybe (Name, Name, Name)
 unboxedRepNames ty
