@@ -461,11 +461,7 @@ reifyDataInfo name = do
  `recover`
   do DatatypeInfo { datatypeContext   = ctxt
                   , datatypeName      = parentName
-#if MIN_VERSION_th_abstraction(0,3,0)
                   , datatypeInstTypes = tys
-#else
-                  , datatypeVars      = tys
-#endif
                   , datatypeVariant   = variant
                   , datatypeCons      = cons
                   } <- reifyDatatype name
